//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "BMP2PPProcs.h"
//---------------------------------------------------------------------------
static void putint (byte *p,int n)
{
 p[0]=(byte)(n&255);
 p[1]=(byte)((n>>8)&255);
 p[2]=(byte)((n>>16)&255);
 p[3]=(byte)((n>>24)&255);
}

int SetTBitmapBits (Graphics::TBitmap *bitmap,
                 int nBitsPerPixel, int nColoursUsed,
                 int nWidthImageBuffer,
                 int nWidthImage, int nHeightImage,
                 byte *pBitmapBits, byte *pPalette)
{
 int i,j,nPadBytes,nImageBytes,nBufBytes;
 static byte bitmapheader[]=
 {
  /* BITMAPFILEHEADER */
  0x42,0x4D,               /* 0 type='BM'                                   */
  0x00,0x00,0x00,0x00,     /* 2 size of file in bytes                       */
  0x00,0x00,               /* 6 reserved                                    */
  0x00,0x00,               /* 8 reserved                                    */
  0x00,0x00,0x00,0x00,     /* 10 offset of bitmap data                      */
  /* BITMAPINFOHEADER */
  0x28,0x00,0x00,0x00,     /* 14 size of this structure in bytes            */
  0x00,0x00,0x00,0x00,     /* 18 width of the image in pixels               */
  0x00,0x00,0x00,0x00,     /* 22 height of the image in pixels              */
  0x01,0x00,               /* 26 number of planes                           */
  0x00,0x00,               /* 28 bits per pixel                             */
  0x00,0x00,0x00,0x00,     /* 30 compression method                         */
  0x00,0x00,0x00,0x00,     /* 34 size of the image in bytes                 */
                           /* set to 0 if no compression is used            */
  0xE8,0x05,0x00,0x00,     /* 38 horizontal resolution in pixels per meter  */
  0xE8,0x05,0x00,0x00,     /* 42 vertical resolution in pixels per meter    */
  0x00,0x00,0x00,0x00,     /* 46 number of colours used in the colour table */
  0x00,0x00,0x00,0x00      /* 50 number of colours used in the image        */
 };
 TMemoryStream *memoryStream=new TMemoryStream;
 bitmapheader[28]=(byte)nBitsPerPixel;
 putint (bitmapheader+46,nColoursUsed);
 putint (bitmapheader+50,nColoursUsed);
 putint (bitmapheader+18,nWidthImage);
 putint (bitmapheader+22,nHeightImage);
 putint (bitmapheader+10,sizeof(bitmapheader)+nColoursUsed*4);
 nBufBytes=(nWidthImageBuffer*nBitsPerPixel+7)/8;
 i=nImageBytes=(nWidthImage*nBitsPerPixel+7)/8;
 nPadBytes=(4-(i&3))&3;
 i+=nPadBytes;
 putint (bitmapheader+2,(i*nHeightImage)+sizeof(bitmapheader)+nColoursUsed*4);
 memoryStream->Write (bitmapheader,sizeof(bitmapheader));
 for (i=0;i<nColoursUsed;++i)
 {
  memoryStream->Write (pPalette+i*3+2,1);
  memoryStream->Write (pPalette+i*3+1,1);
  memoryStream->Write (pPalette+i*3+0,1);
  byte tmp=0;
  memoryStream->Write (&tmp,1);
 }
 for (i=nHeightImage-1;i>=0;--i)
 {
  memoryStream->Write (pBitmapBits+nBufBytes*i,nImageBytes);
  for (j=0;j<nPadBytes;++j)
  {
   byte tmp=0;
   memoryStream->Write (&tmp,1);
  }
 }
 memoryStream->Seek (0,soFromBeginning);
 bitmap->LoadFromStream (memoryStream);
 delete memoryStream;
 return 0;
}

static short bytes2short (byte *p)
{
 return (short) (((int)p[0]) + (((int)p[1])<<8));
}

static int bytes2int (byte *p)
{
 return p[0]+(p[1]<<8)+(p[2]<<16)+(p[3]<<24);
}

int GetTBitmapBits (Graphics::TBitmap *bitmap,
                int *nWidth,int *nHeight,
                byte **pBitmapBytes)
{
 int i,j,k,nBytesPerRow,nBitsPerPixel;
 byte bitmapheader[54],palette[768],*rowbuf,*p;
 if (bitmap->Empty) return -100;
 TMemoryStream *memoryStream=new TMemoryStream;
 bitmap->SaveToStream (memoryStream);
 memoryStream->Seek (0,soFromBeginning);
 /* Read the header */
 memoryStream->Read (bitmapheader,sizeof(bitmapheader));
 /* Check file type */
 if (bitmapheader[0]!='B' || bitmapheader[1]!='M')
 {
  delete memoryStream;
  return -1;
 }
 /* Check number of planes */
 if (bytes2short(bitmapheader+26)!=1)
 {
  delete memoryStream;
  return -2;
 }
 /* Check compression method */
 if (bytes2int(bitmapheader+30))
 {
  delete memoryStream;
  return -3;
 }
 *nWidth=bytes2int(bitmapheader+18);
 *nHeight=bytes2int(bitmapheader+22);
 nBitsPerPixel=bytes2short(bitmapheader+28);
 /* Check for valid width, height, bits per pixel */
 if (nWidth<=0 || nHeight<=0 ||
     (nBitsPerPixel!=1 && nBitsPerPixel!=4 && nBitsPerPixel!=8 &&
      nBitsPerPixel!=24))
 {
  delete memoryStream;
  return -4;
 }
 /* Read in palette */
 if (nBitsPerPixel<=8)
 {
  memoryStream->Seek (bytes2int(bitmapheader+14)+14,soFromBeginning);
  memset (palette,0,768);
  j=bytes2int (bitmapheader+46);
  if (j>256) j=256;
  if (j==0) j=1<<nBitsPerPixel;
  for (i=0;i<j;++i)
  {
   memoryStream->Read (palette+i*3+2,1);
   memoryStream->Read (palette+i*3+1,1);
   memoryStream->Read (palette+i*3+0,1);
   byte tmp;
   memoryStream->Read (&tmp,1);
  }
 }
 /* Seek to bitmap data */
 memoryStream->Seek (bytes2int(bitmapheader+10),soFromBeginning);
 /* Get number of bytes per row in bitmap file and allocate row buffer */
 nBytesPerRow=(*nWidth)*nBitsPerPixel/8;
 if (nBytesPerRow&3) nBytesPerRow=(nBytesPerRow&(~3))+4;
 rowbuf=new byte[nBytesPerRow];
 /* Allocate bitmap buffer */
 *pBitmapBytes=new byte[(*nWidth)*(*nHeight)*3];
 /* Read in file */
 p=*pBitmapBytes+(*nWidth)*(*nHeight-1)*3;
 for (i=0;i<*nHeight;++i,p-=2*(*nWidth)*3)
 {
  memoryStream->Read (rowbuf,nBytesPerRow);
  for (j=0;j<*nWidth;++j,p+=3)
   switch (nBitsPerPixel)
   {
    case 24: p[0]=rowbuf[j*3+0];
             p[1]=rowbuf[j*3+1];
             p[2]=rowbuf[j*3+2];
             break;
    case 8:  k=rowbuf[j];
             p[2]=palette[k*3+0];
             p[1]=palette[k*3+1];
             p[0]=palette[k*3+2];
             break;
    case 4:  k=rowbuf[j/2]; if (j&1) k&=15; else k>>=4;
             p[2]=palette[k*3+0];
             p[1]=palette[k*3+1];
             p[0]=palette[k*3+2];
             break;
    case 1:  k=(rowbuf[j/8]>>(7-(j&7)))&1;
             p[2]=palette[k*3+0];
             p[1]=palette[k*3+1];
             p[0]=palette[k*3+2];
             break;
   }
 }
 delete[] rowbuf;
 delete memoryStream;
 return 0;
}